*	(C)1990, Lindley Systems

;		By Robert Lindley and William Lindley

;

;	PRINTER DRIVER WITH USER SELECTABLE FEATURES FOR CP/M

;	V4.2  24-JUL-82 CP/M VERSION WITH SETTABLE FEATURES

;	V4.3  15-AUG-82 ADDED 8255 PORT CODE

;	V4.3a 25-AUG-82 "SET" CONDITIONAL ASSEMBLY

;	V4.4  27-AUG-82 ADDED Z89-11 PARALLEL CODE

;	V4.4a 14-NOV-82 COMMON I/O WITH NEC DRIVER

;	V4.4b 15-JAN-83 FIX GRAPHICS COUNT DOWN

;	V4.4c 24-JUL-83 ADDED RZ AFTER CPI '-'

;	V4.5  11-OCT-83 IMPROVED INSTALLATION METHOD

;	V5.0  13-MAR-84 ALL NEW INSTALLATION METHOD

;			+ STAR GRAPHICS

;	V5.1  08-JUL-84 H8 SIGNON FIX

;	V5.2  17-SEP-84 H89PIP CODE ADDED

;

;	V6.0  24-AUG-90 All new code (from HDOS Version 6.0 driver) WL



FALSE	EQU	0

TRUE	EQU	1

;	Printer Type:

MX	EQU	TRUE	; Epson & Compatibles

NEC	EQU	FALSE	; NEC-8023, C. Itoh 8510

DP	EQU	FALSE	; Dataproducts (Paper Tiger)

MPI	EQU	FALSE	; MPI

OKI	EQU	FALSE	; Okidata (Microline)

H25	EQU	FALSE	; H-25

PLAIN	EQU	FALSE	; 'Plain' printer (Daisywheel, TTY, etc.)

DIABLO	EQU	FALSE	; Diablo/Compatible daisywheel

;	NOGRAF is set to TRUE only for original MX-80's and

;	non-Epson compatible printers.	This causes the *driver*

;	to interpret ESC > and ESC = to turn the high bit on and off.

;	All other Epson-type printers understand these codes.

NOGRAF	EQU	FALSE	; (TRUE for Original MX-80)

;	Set FFSIM to TRUE if you want to simulate FF's with NL's.

FFSIM	EQU	FALSE

;	Number of Printer Units and Macros:

MAXMAC	EQU	4

;	Port Type:

U8250	EQU	TRUE	; H/Z-89 SERIAL, H8-4

Z8255	EQU	FALSE	; H-89-11 OR H&H PARALLEL PORT

U8251	EQU	FALSE	; H8-5, H8-2

U8255	EQU	FALSE	; TRUDATA PARALLEL

DIRECT	EQU	FALSE	; DIRECT PORT I/O

PERCOM	EQU	FALSE	; Z controller

WIDGET	EQU	FALSE	; MicroWidget

H89PIP	EQU	FALSE	; FBE H89PIP

;

TAB	EQU	9	; Tab

FF	EQU	12	; Form Feed

CR	EQU	13	; Carriage Return

ENL	EQU	212Q	; Ending New Line

NL	EQU	012Q	; New Line

ESC	EQU	27	; Escape

SPACE	EQU	32

;

; BIOS INTERFACE HEADER

	DW	PARAMS	; POINTER TO PARAMETER TABLE

RTSAD:	DW	RTSAD	; ADDRESS TO BE RESTORED

	JMP	SIGNON	; OUR SIGNON ROUTINE

	JMP	LPTOS	; CHECK STATUS ENTRY

; BEGINNING OF DRIVER CODE

;	 

; WRITE DATA TO THE DEVICE.

;

;	 (C) = CHARACTER TO PRINT

;

; CHARACTER PROCESSING

;	 THE FOLLOWING PROCESSES ONE CHARACTER

;	 IF A SEQUENCE REQUIRES MORE THAN ONE,

;	 A VARIABLE IS SET FOR NEXT STATE.

;

LPTOUT: LDA	LPINIT

	ORA	A

	CZ	LPTINI

	LXI	D,GSTATE ; (DE) points at GSTATE

	LDAX	D	; Get preprocessing state

	CALL	LPINX	; GO TO STATE

	DB	TRYCCC-$ ; Normal characters (State 0)

	DB	GSTATE1-$ ; LSB

	DB	GSTATE2-$ ; MSB

	DB	GSTATE3-$ ; COUNT DOWN

 IF MX

	DB	GSTATE4-$ ; Get LSB

	DB	GSTATE5-$ ; Get MSB

 ENDIF

 IF NEC

	DB	GSTATE4-$ ; Get LSB

	DB	GSTATE4-$

	DB	GSTATE4-$

	DB	GSTATE4-$ ; Get MSB

 ENDIF

;

;	Graphics States:

;

 IF MX

;

; GET COUNT LSB

GSTATE4: MOV	A,C

	STA	GRLSB

GS4A:	XCHG		; H = GSTATE

	MOV	D,A

	DCR	M

	JMP	SENDNC



; GET COUNT MSB

GSTATE5: MOV	A,C

	STA	GRMSB

	JMP	GS4A	; Decrement state & send character

 ENDIF

 IF NEC

; Get Graphics count digits (Then begin counting down)

GSTATE4: LXI	H,GRMSB

	PUSH	B	; Char. in C

	PUSH	D	; (D->GSTATE)

	CALL	LPDIG2	; Char. in C

	POP	D

	POP	H	; H -> GSTATE

	DCR	M	; Count down states

	MOV	D,C

	JMP	SENDNC	; Send character to printer

 ENDIF

 IF NEC OR MX

;

; Sending graphics. Count down count until 0.

;

GSTATE3: MOV	D,C	; Char in	D

	LHLD	GRLSB	; Get 16 bit count

	MOV	A,L	; Check if HL=0

	ORA	H

	DCX	H	; Decrement graphics count

	SHLD	GRLSB

	JNZ	SENDNC	; Count = 0? If not, send char. unchanged.



;	Graphics counted down to zero. Process this char normally.



	XRA	A

	STA	GSTATE

	JMP	TRYCCC

 ENDIF

 IF OKI OR DP

;

;	Sending graphics, terminated with Ctrl-C, Ctrl-B.

;	Printer treats Ctrl-C, Ctrl-C as 'quoted' 3.

;

GSTATE3: MOV	A,C

	CPI	3	; Find a ctrl-C?

	JZ	GS3A	; Yes, process.

	CPI	2	; Find ctrl-B?

	JZ	GS3B



	XRA	A

	STA	GRMSB	; Zero "ctrl-C found" flag.

GS3SEND: MOV	D,C

	JMP	SENDNC	; Send character



; Got a Ctrl-C:

GS3A:	LDA	GRMSB	; First Ctrl-C sets GRMSG to -1;

	CMA		; then next Ctrl-C sets to 0 (printer quoted ^C);

	STA	GRMSB	; Ctrl-B (below) or any other char also sets to 0.

	JMP	GS3SEND



; Got Ctrl-B:

GS3B:	LDA	GRMSB

	ORA	A

	JZ	GS3SEND	; not after Ctrl-C, send normally

; Ctrl-C then Ctrl-B -- exit graphics mode after this character

	XRA	A

	STA	GSTATE

	JMP	GS3SEND

 ENDIF

 IF MPI

;

;	Graphics data must have all bit-7's TRUE.

;	The printer drops out of graphics mode with terminating

;	character whose 7th bit is 0.

;

GSTATE3: MOV	A,C

	ANI	64	; Bit 7 TRUE if graphics data

	MOV	D,C	; Ready to call SENDNC

	JNZ	SENDNC	; 7th bit true, stay in graphics mode

	XRA	A	; 7th bit false, reset to normal state after this char.

	STA	GSTATE

	JMP	SENDNC

 ENDIF

 IF H25

;

;	This handles escape codes which begin with 'Esc-['. Those codes

;	may have many characters, but are always terminated with

;	a character >= 'A'.

;

GSTATE3: MOV	A,C

	CPI	'A'

	JC	SENDNC	; Less than 'A', keep looking for end of escape sequence

	XRA	A	; If found 'A' or above, send terminating character

	STA	GSTATE	; and go back to state 0.

	JMP	SENDNC

 ENDIF

 IF DIABLO OR PLAIN

;	Process character normally

GSTATE3: XRA	A

	STA	GSTATE

	JMP	TRYCCC

 ENDIF



;

; Macro	Reprogramming (Get Macro #)

;

GSTATE2	XRA	A	;A = 0

	STAX	D	;GSTATE = 0. Process next char normally.

; Initiate macro programming.

	MOV	A,C

	CALL	MAC1	;Find addr in HL, macro # in A

	SHLD	PROGADR	;Save programming address



	MVI	A,1	;On first character of macro.

	STA	PROGMAC	;Save flag



	ORA	A	;Clear carry (show no Ctrl-C detected)

	RET



; User Set programming.

;	Invoked with "~*" and continuing through next "*".

;	Special characters select Formlength, Pagelength, Macroinvoke,

;	Tab setting, Leftmargin.

;

GSTATE1: MOV	A,C

	CPI	'0'

	JC	GS1XIT	; Less than '0', terminate.

	CPI	'9'+1

	JNC	GS1C	; Greater than '9', see if command.

; Accumulate number:

	LHLD	PROGADR ; Address of variable

	MOV	A,H

	ORA	L	; See if HL=0

	CNZ	LPDIG	; If HL<>0, accumulate number:

;			; (Addr in HL, char in C)

	XRA	A	; Clear A, flags

	RET



GS1C:	PUSH	B

	CALL	GSSENDM

	POP	B



GS1CMD:	LXI	H,FLEN	; First user-set variable

	MOV	A,C

	MVI	B,0	; Default reset-value to save

	ORI	20H	; Make it lower-case



	CPI	'f'	; Formlength

	JZ	GS1OK

	INX	H

	CPI	'p'	; Pagelength

	JZ	GS1OK

	INX	H

	CPI	'l'	; Leftmargin

	JZ	GS1OK

	INX	H

	CPI	't'	; Tabs

	JZ	GS1OK

	INX	H

	CPI	'm'

	JNZ	GS1XIT	; Unknown, exit.

	MVI	B,8	; Set macro # to 8 so we know we're invoking

;			; (variable will be 80+actual macro #, gets masked)

GS1OK:	SHLD	PROGADR ; Save Programming address

	MOV	M,B	; Save the Reset-Value

	XRA	A

	RET



GS1XIT:	XRA	A

	STAX	D	; GSTATE = 0. Process next normally

	CALL	GSSENDM ; Send macro if invoked

	CALL	LPSPECI ; reset form, page, tabs...

	RET



; If we invoked macro, send it.

GSSENDM LXI	H,PROGMAC

	MOV	A,M	; Get macro # if any

;	If calling macro, Flag contains 8n (hex) where n=0..7 for macro #

	ORA	A	; Were we invoking a macro?

	RZ		; no, do nothing

	MVI	M,0	; Reset flag

	CALL	MAC1	; Get address in (HL)

	CALL	SENDMAC	; Send it

	RET



;	(STATE 0)

;

;	Process	potential Control Conversions here

;	Input: Character in C

;

;	Three special sequences	after a	control	conversion are defined:

;	@, *, and #. @ removes the next	Newline; # redefines a macro;

;	'*' invokes User settings (of form/page length, macro#, tab columns,

;	left margin)



TRYCCC:	LXI	D,CCCSW	; Ctrl conversion switch (true if previously found tilde)

	LDAX	D

	ORA	A	; Is switch on?

	MOV	A,C	; [Character in A]

	JNZ	INCCC	; Convert



; Test for new ctrl conversion

GTC2:	LXI	H,CCC	; Check current char against CCC.

	CMP	M

	JNZ	OUTPUT	; No control conversion -- send char [in C]



	XCHG		; (HL) points to CCCSW

	INR	M	; Switch on: convert next char

	RET



; Ctrl conversion in progress

INCCC:	XCHG		; (HL) = CCCSW

	MVI	M,0	; Reset CCCSW for next time



	LXI	H,CCC	; Point to conv char

	CMP	M	; Look for two ccc's

	JZ	OUTPUT	; If so, send ONE.

; Special control conversions: ~@, ~*, ~#

	CPI	'@'	; @ MEANS SKIP NEXT NL

	JZ	KILLNL

	CPI	'*'

	JZ	INVOKE

	CPI	'#'

	JZ	REPROG

; Perform control conversion

	ANI	037Q

	MOV	C,A

	JMP	OUTPUT



; New Line skip

KILLNL:	LXI	H,NLSW	; SET TO KILL NEXT NL

	MVI	M,1

	RET

; Start user settings

INVOKE:	XRA	A	; Show we're not invoking macro

	STA	PROGMAC

	LXI	H,0	; Set PROGADR=0 to show no variable yet

	SHLD	PROGADR

	INR	A	; Graphics State 1, Start User Settings

	STA	GSTATE

	RET



; Macro	reprogramming

REPROG:	LDA	PROGMAC	; Are we programming now?

	ORA	A

	JNZ	REPROG1



; Initiate reprogramming:

	MVI	A,2	; State 5 -- Get macro reprogramming #

	STA	GSTATE

	RET



; Terminate reprogramming. Fill	rest of	macro with zeros.

REPROG1:	LHLD	PROGADR	; Get programming address

	LDA	PROGMAC		; Get count of characters processed+1

REPROG2:	CPI	17

	JNC	ENDPROG

	MVI	M,0		; Store zero

	INX	H

	INR	A

	JMP	REPROG2

; Reset	states:

ENDPROG:	XRA	A

	STA	PROGMAC

	STA	GSTATE

	RET

;

;	State machine for escape code interpretation.

;	These states are entered AFTER control conversion.

;

;	Input: Character in C (Moved to	D)

;

OUTPUT:	MOV	D,C	; Character into D

	LXI	B,STATE	; Get state variable

	LDAX	B



; At entry to each state:

;	(BC) points to STATE

;	D contains character

;

	CALL	LPINX	; Go to state for:

	DB	STATE0-$	; Normal printing

	DB	STATE1-$	; Start Escape sequence

	DB	STATE2-$	; Pass until Null

	DB	STATE3-$	; ESC with one argument, not masked

	DB	STATE4-$	; ESC with one argument, masked to 0..15

	DB	STATE5-$	; ESC with two arguments, masked 0..15

;

;  STATE 1 - Process first character after an Escape character.

;

STATE1:	XRA	A

	STAX	B	; State 0

	MOV	A,D	; Char in A

 IF NOGRAF

; Test for set high bit	code (Original MX-80's and non-Epsons)

	CPI	'>'

	JNZ	S1A

	MVI	A,80H	; HIGH BIT ON

	JMP	S1XX

; Test for reset high bit code

S1A:	CPI	'='

	JZ	S1XW

	CPI	'#'

	JNZ	S1B

S1XW:	XRA	A	; HIGH BIT OFF

S1XX:	STA	HIBIT

	RET

 ENDIF

S1B:	MVI	A,27	; SEND THE ESCAPE CODE

	CALL	OUTCHAR

	CALL	ESCTYP

	JMP	SENDNC

;

; State	2 - Send Esc sequence until NULL

;

STATE2:	XRA	A	; A=0

	ORA	D	; Is D=A (Is D=0?)

	JNZ	SENDNC	; If we find a NULL, fall through...

;

; State	3 - One	character after	ESC, not masked

;

STATE3:	JMP	SENDNC0	; Send character & reset to state 0



;

; State	4 - Mask character after ESC

;

STATE4:	XRA	A	; New state = 0

	JMP	MASKIT

;

; State	5 - Mask TWO characters	after ESC

;	(Go through State 5, then State	4)

;

STATE5:	LDAX	B	; Decrement State (to 4)

	DCR	A

MASKIT:	STAX	B	; Store new state

	MVI	A,17Q	; Mask to be 0..15

	ANA	D

	MOV	D,A	; Save char in D

	JMP	SENDNC	; Send first character



;*************************************************

;

;  State 0 - Normal character processing

;

STATE0:	MOV	A,D	; Get character

	ANI	177Q	; Strip top bit for testing

	CPI	SPACE	; Test for control char

	JNC	SENDC	; Printing character -- send & incr. col. count



; Control Character Processing



GOTCTRL: CPI	ESC	; Escape?

	JZ	GOTESC

; All except ESC - if programming macro, save code as is.

	LDA	PROGMAC

	ORA	A

	JNZ	SENDNC



	MOV	A,D

	CPI	CR	; Carriage return?

	JZ	GOTCR

	CPI	NL	; Newline?

	JZ	GOTNL

	CPI	FF	; Form feed?

	JZ	DOFF

	CPI	TAB	; Tab?

	JZ	GOTTAB

 IF OKI OR DP

	CPI	3

	JZ	GOTGRAF

 ENDIF

	JMP	SENDNC	; Some other code, send w/o column count



;

; Upon receiving an ESC	code, the state	is set to 1 for	further	processing.

;

GOTESC:	MVI	A,1	; Set STATE = 1

	STAX	B

	RET		; Send nothing yet (State 1 sends the ESC)

;

; Carriage Return

;

GOTCR:	LDA	NLSW	; SKIP?	V4.1A

	ORA	A	; V4.1A

	RNZ		; V4.1A

	MVI	A,1

	STA	CX	; Back to column 1

	JMP	SENDNC



; Process Newlines.

GOTNL:	LXI	H,NLSW	; TEST FOR SKIP NL

	MOV	A,M

	ORA	A; 

	MVI	M,0	; RESET NLSW

; Skip newline

	RNZ

; Process newline

	INR	A	; A=1 (was 0)

	STA	CX	; Column = 1

	LDA	PLEN	; PAGE LENGTH

	ORA	A	; DON'T PROCESS IF ZERO

	JZ	DONL

	LXI	H,LX

	INR	M	; SET NEXT LINE

	CMP	M

	JC	DOFF	; Past end of page? Send formfeed

;

; DONL -- Send a CR/LF pair

;

DONL:	MVI	A,CR

	CALL	OUTCHAR

	MVI	D,NL

	JMP	SENDNC

;

; Do a form feed

;

DOFF:	MVI	A,1	; Reset to Col 1 Line 1

	STA	CX

	STA	LX

	MVI	A,CR	; Send Carriage return too

	CALL	OUTCHAR

 IF FFSIM		; Simulate FF via NL's

FFLP:	LXI	H,LX

	INR	M

	LDA	FLEN

	CMP	M

	RC	FFXIT	; Return if done

	MVI	A,NL

	CALL	OUTCHAR

	JMP	FFLP

 ELSE			; Printer understands FF's

	MVI	D,FF

	JMP	SENDNC

 ENDIF



;

; Tab Processing

;

GOTTAB:	CALL	COL1	; Test for col 1

	LDA	TABSW

	ORA	A

	JZ	SENDNC	; Not processing tabs? Send them "straight."

	LDA	CX

	MOV	B,A	; Column # in B (1..n)

	LDA	TABSW

	MOV	D,A	; # of Columns / Tab in D

TAB0:	MOV	C,B	; Column # in C

	MOV	A,D

	INR	A	; TABSW + 1 is desired column modulo

	SUB	C	; A = TABSW - COLUMN

	JZ	TAB1	; If already on tab stop, move over one tab.

	JNC	TAB2	; No carry means column # was < TABSW

	MOV	A,C	; A = column #

	SUB	D	; column # - TABSW

	MOV	B,A	; B is temporary copy of CX

	JMP	TAB0



TAB1:	MOV	A,D	; Move over an entire tab (D=TABSW)

TAB2:	MOV	C,A	; C = # of spaces to put out

	ADD	B	; Add to current column # to get Destination column

	STA	CX

TABLOOP: MVI	A,SPACE

	CALL	OUTCHAR

	DCR	C

	JNZ	TABLOOP

	RET



 IF OKI OR DP

;

;	Okidata - Ctrl-C launches us into graphics mode

;

GOTGRAF: EQU *

	MOV	D,A	; Save character

	MVI	A,3	; Graphics 'countdown'

	STA	GSTATE

	JMP	SENDNC

 ENDIF

;

; ROUTINE TO PROCESS COLUMN ONE

;

; USES: A,C

COL1:	LDA	CX	; COLUMN COUNTER

	DCR	A	; SEE IF A = 1

	RNZ		; COL 1 MEANS STARTING NEW LINE

	LDA	LMVAL

	INR	A	; 

	MOV	C,A

LMLOOP:	DCR	C	; PUT SPACES AT LEFT MARGIN

	JZ	TRYXC

	MVI	A,SPACE	; CX IS NOT INCREMENTED

	CALL	OUTCHAR

	JMP	LMLOOP

;

; EXTSW, if low	bit is true, tells us to expand	each line.

;

TRYXC:	LDA	EXTSW

	ANI	1	; Low bit true?

	RZ

 IF H25 OR MX OR NEC

	MVI	A,14	; Ctrl-N = Elongated

	CALL	OUTCHAR

 ENDIF

 IF MPI

	MVI	A,27

	CALL	OUTCHAR

	MVI	A,14

	CALL	OUTCHAR

 ENDIF

 IF DP

	MVI	A,1	; Ctrl-A = Elongated

	CALL	OUTCHAR

 ENDIF

 IF OKI

	MVI	A,31	; Ctrl-_ = Elongated

	CALL	OUTCHAR

 ENDIF

	RET



;*************************************************************************

;

;	MACRO ROUTINES:

;

; Find & save address of macro

;	INPUT:	Macro #	(0..7) in A

;	OUTPUT:	Address	of macro in (HL)

;		Macro #	(0..MAXMAC-1) in A

;

MAC1:	ANI	7	; Allow 0..7 only

	CPI	MAXMAC+1

	JC	MAC2	; > MAXMAC?

	XRA	A	; Set to 0



MAC2:	LXI	H,ODAT	; Macro data area

	PUSH	PSW

	ADD	A	; Mult A by 16

	ADD	A

	ADD	A

	ADD	A

	MOV	E,A	; UNIT * 16

	MVI	D,0

	DAD	D

	POP	PSW

	RET

;

; Send Macro at	(HL)

;

SENDMAC: MOV	A,M	; Get expanded char switch

	STA	EXTSW

	MVI	E,15	; 15 other characters

SENDM1:	INX	H

	MOV	A,M

	CALL	OUTCHAR

	DCR	E

	JNZ	SENDM1

	RET

;

; Routine to program macros.

; Macro	# is in	A (1..8) and character to save is in D.

;

; Increment count of characters	processed; check for overflow.

DOPROG:	LXI	H,PROGMAC

	MOV	A,M

	CPI	17	; Past end of macro buffer?

	RZ		; if so, ignore character.

	INR	A

	MOV	M,A

; Save character in macro buffer

	LHLD	PROGADR	; Get address

	MOV	M,D	; Save character

	INX	H

	SHLD	PROGADR



	ORA	A	; Clear carry (show no Ctrl-C detected)

	RET

;

;*************************************************

;

; Classify Escape type,	and store in state variables (STATE and GSTATE)

;

;	Character is in	D; count state #'s in B

;

ESCTYP:	LXI	H,ESCTBL

	MVI	B,1

;

;	Type 1 -- Graphics (printer specific) -- Sets GSTATE variable

;	Type 2 -- Send until NUL

;	Type 3 -- send 2nd character, not masked

;	Type 4 -- send 2nd character, masked to	be 0..15

;		(for escape codes needing <0> or <1> sent)

;

ESCTYP0: MOV	A,M		; Found matching code?

	CMP	D

	JZ	ESCTYP1		; Found escape type, save state var.



	INX	H		; Go to next character

	ORA	A		; Was that the end of this part of table?

	JNZ	ESCTYP0



; (HL) now points to first char	of next	part of	table

; Advance to next state, stop if end of	table



	INR	B		; Count up states



	MOV	A,M		; Is there another state table?

	ORA	A

	JNZ	ESCTYP0		; Found end of state table



	MOV	B,A		; Unclassified character (A=0)



ESCTYP1:	MOV	A,B	; Get escape type

	CPI	1

	JNZ	ESCTYP2		; Types other than 1 go into STATE



; Graphics: Set GSTATE -> Get MSB count

 IF MX

	MVI	A,5

 ENDIF

 IF NEC

	MVI	A,7

 ENDIF

; If neither NEC or MX:

 IF DP OR MPI OR OKI OR H25 OR PLAIN OR DIABLO

	MVI	A,3

 ENDIF

	STA	GSTATE

	XRA	A		; STATE = 0 (normal when done)



ESCTYP2: STA	STATE		; Save new STATE

	RET



;

; Get digits for user settings

;

;   (H,L) -> 1-BYTE ACCUMULATOR LOCATION

;   (C)      NEXT DIGIT CHARACTER

;

; USES A

LPDIG:	PUSH	B

	MOV	A,C	; Make C [0..15]

	ANI	0FH

	MOV	C,A

	MOV	A,M	; FETCH CURRENT VALUE

	MOV	B,A	; Also put in B

	ADD	A	; * 2

	ADD	A	; * 4

	ADD	B	; * 5

	ADD	A	; * 10

	ADD	C	; Add character

	MOV	M,A	; RETURN UPDATED VALUE

	POP	B

	RET



 IF NEC

;

; Get digits for graphics length

;

;   (H,L) -> 2-BYTE ACCUMULATOR LOCATION

;   (C)      NEXT DIGIT CHARACTER

;

; USES A,D,E,H,L

LPDIG2:	PUSH	H

	MOV	E,M	; Fetch current value

	INX	H

	MOV	D,M

	MOV	L,E

	MOV	H,D

	DAD	H	; * 2

	DAD	H	; * 4

	DAD	D	; * 5

	DAD	H	; * 10

	MOV	A,C	; Character value

	ANI	0FH

	ADD	L	; Add it on

	MOV	L,A

	JNC	LPDIG2A	; Do we have to carry?

	INR	H

; Value in HL:

LPDIG2A: POP	D	; Get address

	XCHG		; Value in DE, addr in HL:

	MOV	M,E

	INX	H

	MOV	M,D

	RET

 ENDIF

;

;*********************************************

;

;	LPSPECI	Printer Specific Initialization

;

;	When we use the User-Set feature we come here to reset

;	form len/page len/tabs/etc.

;

LPSPECI:

 IF MX

	LDA	FLEN	; Get form length

	ORA	A

	RZ		; Don't initialize if zero

	MVI	A,ESC	; Set form length

	CALL	OUTCHAR

	MVI	A,'C'

	CALL	OUTCHAR

	LDA	FLEN

	CALL	OUTCHAR

 ENDIF

	RET		; (MX falls through; all others but NEC return)

 IF NEC

;

;	Set Horizontal Tab sequence

;

	CALL	LPPNXT

	DB	27,'('		;START TAB SET SEQUENCE

	DB	'009,017,025,033,041,049,057,065,073.',0

;

;	Set Vertical Format Unit

;

	LXI	H,FLEN	;FORM LENGTH

	MOV	A,M

	DCR	A

	RZ		;USE PRINTER DEFAULT VFU

	RM

	MOV	E,A	;LOOP COUNT

; START VFU SEQUENCE

	CALL	LPPNXT

	DB	1DH,41H,40H,0

; LOOP TO FORM LENGTH

LPVFU1: DCR	E	;END OF FORM

	JZ	LPVFU3

	CALL	LPPNXT

	DB	40H,40H,0 ; NORMAL FILL

	JMP	LPVFU1

; END VFU SEQUENCE

LPVFU3: CALL	LPPNXT

	DB	43H,40H,41H,40H,1EH,0

	RET

;

; SEND INLINE CODES TO PRINTER, STOP ON NULL

LPPNXT: XTHL		;(H,L)->TEXT

	CALL	LPPRLP

	XTHL		;(STACK)->RETURN ADDR

	RET

; SEND (H,L)->CODES TO PRINTER

LPPRLP: MOV	A,M	;GET CHAR

	INX	H	;POINT TO NEXT CHAR

	ORA	A	;SET FLAGS

	RZ		;DO NOT SEND NULL

	CALL	OUTCHAR

	RC

	JMP	LPPRLP	; DO IT AGAIN

 ENDIF



;

LPOUT:	PUSH	PSW

	PUSH	H

	LXI	H,LPPORT

	ADD	M

	POP	H

	STA	LPO1+1

	MOV	A,C

LPO1:	OUT	0

	POP	PSW

	RET





;*************************************************

;

;	CHARACTER SENDING ROUTINES

;



;  Send	char in	D with no column count,	and reset state	to 0.

;

SENDNC0: XRA	A

	STA	STATE

	JMP	SENDNC

;

;  Send	char in	D with column count

;

SENDC:	CALL	COL1	; TEST FOR COL 1

	LXI	H,CX	; COLUMN COUNTER

	INR	M	; INCREMENT COLUMN COUNT

;

; Send character in D -- No column count

;	With hibit setting if enabled.

;

SENDNC:	LDA	PROGMAC	; Are we programming a macro?

	ORA	A

	JNZ	DOPROG	; Yes -- save character



	MOV	A,D	; Get character in A

 IF NOGRAF

	LXI	H,HIBIT	; Set high bit accordingly

	ORA	M

 ENDIF

;

	CALL	OUTCHAR	; Sends character in A

	RET

;********************************************

;

;	OUTCHAR	Send character to printer

;	(Because some routines use C as a counter)

;

;	Input:	A	Character to send

;	Output:	'C'	Carry flag set if error, clear if no error

;	Uses:	A

OUTCHAR: PUSH	B

	MOV	C,A

	CALL	LPCHAR

	POP	B

	RET

;

; LPCHAR - CHARACTER IN C

;	     TO LPT PORT

LPCHAR: CALL	LPTOSX

	ORA	A

	JZ	LPCHAR	; 0 => PRINTER BUSY

	XRA	A

 IF H89PIP

	INR	A	; ONE

 ENDIF

 IF WIDGET

	INR	A	; ONE

 ENDIF

	CALL	LPOUT

 IF Z8255

	MVI	A,3

	MVI	C,000H	; STROBE LOW - ASSERTED

	CALL	LPOUT

	MVI	C,001H	; STROBE HIGH - INACTIVE

	CALL	LPOUT

 ENDIF

 IF WIDGET

	MVI	A,12

	CALL	LPIN	; GET STATUS

	XRI	008H	; FLIP CLOCK BIT

	ORI	010H	; STROBE HIGH - ASSERTED

	MOV	C,A

	MVI	A,12

	CALL	LPOUT

	CALL	LPIN	; GET STATUS

	XRI	008H

	ANI	0EFH	; STROBE LOW  - INACTIVE

	MOV	C,A

	MVI	A,12

	CALL	LPOUT

 ENDIF

 IF H89PIP

	MVI	C,060H	; STROBE OUT CHARACTER

	MVI	A,2

	CALL	LPOUT

	MVI	C,070H	; RESET STROBE

	CALL	LPOUT

 ENDIF

	RET

;

LPTOS:	LDA	LPINIT

	ORA	A

	CZ	LPTINI

;

; TEST FOR PRINTER BUSY

LPTOSX: PUSH	H

	PUSH	D

	PUSH	B

 IF U8250

	MVI	A,005Q	; (L) = LINE STATUS REGISTER

	CALL	LPIN

	ANI	060H	; ARE TRANS HOLDING REGISTER

	XRI	060H	;   AND TRANS SHIFT REGISTER

	JNZ	LPNRDY	;     BOTH EMPTY?

	MVI	A,006Q	; MODEM STATUS REGISTER

 ENDIF

 IF U8251

	MVI	A,001Q	; 8251 STATUS

 ENDIF

 IF U8255 OR Z8255

	MVI	A,002Q	; PORT C

 ENDIF

 IF DIRECT

	MVI	A,0

 ENDIF

 IF PERCOM

	MVI	A,5

 ENDIF

 IF WIDGET

	MVI	A,12

 ENDIF

 IF H89PIP

	MVI	A,2	; PORT C

 ENDIF

	CALL	LPIN

	LXI	H,LPMASK

	ANA	M	; ISOLATE BUSY BIT

	INX	H

	XRA	M	; TEST FOR READY MATCH

	JNZ	LPNRDY

	MVI	A,0FFH	; SHOW READY

	JMP	POPRET

LPNRDY: XRA	A	; NOT READY

POPRET: POP	B

	POP	D

	POP	H

	RET

;

; INITIALIZE THE PRINTER

LPTINI: PUSH	H

	PUSH	D

	PUSH	B

	MVI	A,1

	STA	LPINIT	; MARK AS INITIALIZED

;

; 8250 INITIALIZATION

 IF U8250

	MVI	A,1	; INTERRUPT CONTROL

	MVI	C,0	; ALL OFF

	CALL	LPOUT

	MVI	A,3	; LINE CONTROL

	MVI	C,83H	; SET DIVISOR LATCH

	CALL	LPOUT

	MVI	A,4	; MODEM CONTROL

	MVI	C,0FH	; DSR, CTS, RLSD ALL ON

	CALL	LPOUT

	MVI	A,0	; LOW BYTE

	LHLD	LPBAUD

	MOV	C,L

	CALL	LPOUT

	MVI	A,1	; HIGH BYTE

	MOV	C,H

	CALL	LPOUT

	MVI	A,3	; LINE CONTROL

	MVI	C,3	; 8 BIT WORDS, ONE STOP BIT

	CALL	LPOUT	; LOOP OFF

	LXI	H,0

LPDLY:	DCX	H

	MOV	A,H

	ORA	L

	JNZ	LPDLY

 ENDIF

;

; INITIALIZE THE 8251 PORT

; 3 OUTPUTS REQUIRED TO GET TO KNOWN STATE

 IF U8251

	MVI	A,1	; COMMAND REG

	MVI	C,0

	CALL	LPOUT

	CALL	LPOUT

	CALL	LPOUT

; INTERNAL RESET

	MVI	C,100Q	; INTERNAL RESET

	CALL	LPOUT

; MODE BYTE - 014=>8 BITS, 002=>/16, 100=> 1 STOP BIT

	MVI	C,116Q

	CALL	LPOUT

; COMMAND BYTE - 020=>ERROR RESET, 001=>XMIT ENABLE

	MVI	C,021

	CALL	LPOUT

 ENDIF

 IF U8255 OR Z8255

	MVI	A,3	; CONTROL PORT

	MVI	C,250Q	; PORT A, MODE 1 OUTPUT

			; PORT B, MODE 0 OUTPUT

			; PORT C, BITS 4&5 INPUTS

	CALL	LPOUT	;	       0-3 OUTPUTS

 ENDIF

 IF Z8255

	MVI	A,2	; PORT C

	MVI	C,1	; DATA STROBE HIGH - INACTIVE

	CALL	LPOUT

 ENDIF

 IF WIDGET

	MVI	A,3	; CONTROL PORT

	MVI	C,089H

	CALL	LPOUT

 ENDIF

 IF H89PIP

	MVI	A,3	; CONTROL PORT

	MVI	C,081H

	CALL	LPOUT

	MVI	A,2	; PORT C

	MVI	C,070H

	CALL	LPOUT

 ENDIF

	CALL	LPSPECI	; Printer-specific initialization

;

; Clear variables:

	XRA	A

	STA	HIBIT	; High bit

	STA	STATE	; State variable

	STA	GSTATE	; Graphics State variable

	STA	NLSW	; NL switch

	STA	CCCSW	; CCC switch

	STA	PROGMAC ; Not programming macro

;

; Set variables	to 1:

	INR	A

	STA	LX	; Line Counter

	STA	CX	; Column Counter



	JMP	POPRET

;

; TEST FOR BUSY

LPIN:	PUSH	H

	LXI	H,LPPORT

	ADD	M	; ADD OFFSET (A) = PORT

	STA	LPIADD	; PUT IT HERE

LPIADD: EQU	$+1	; HERE'S PORT ADDRESS

	IN	0

	POP	H

	RET



; INDEXED CASE BRANCH

LPINX:	XTHL		; TABLE FOLLOWS CALL

	PUSH	D

	MOV	E,A	; SELECTED BRANCH

	MVI	D,0

	DAD	D

	MOV	E,M	; BRANCH OFFSET

	DAD	D

	POP	D

	XTHL

	RET

;

; PRINTER OPTIONS TABLE

PARAMS: DB	60	; DRIVER VERSION



;

;	Escape Code Classification File

;	For Ultimate Driver Version 6.0

;

;	(c) 1989, 1990 Lindley Systems

;	4257 Berwick Place

;	Woodbridge, VA 22192

;



ESCTBL:	EQU	$



 IF MX

;	Epson/Compatible

;

;	Type 1 -- Graphics

;

        DB	'K'     ; Graphics

        DB	'L'     ; Graphics

	DB	'*'	; Char Set download

	DB	'Y'	; Dual-density, double-speed

        DB	'Z'	; Quad density

; (Newer STAR printers use above codes; older ones use these:)

	DB	'y'	; Dual-density, Double-speed

        DB	'z'	; Quad density



	DB	0	; (end state 1)



;

;	Type 2 -- Send until NUL

;

	DB	'B'	; Horiz tabs

        DB	'D'     ; Vert tabs

	DB	'b'

	DB	0	; (end state 2)

;

;	Type 3 -- send 2nd character, with control conversion,

;	but without changing states

;

	DB	'A'	; Dots/Line

	DB	'C'	; Lines/Form

	DB	'N'	; Skip Page End

	DB	SPACE	; Space - Insert proportional space

	DB	'3'	; Set line spacing, n/216"



	DB	'J'	; n/216" forward feed

	DB	'j'	; n/216" reverse feed

	DB	'I'	; Select character set

	DB	'Q'	; Set right margin

	DB	'N'	; Set bottom margin



	DB	0	; (end state 3)



;

;	Type 4 characters (Mask next byte)

;

	DB	'S'	; Super/Subscript

	DB	'U'	; Uni/Bi-directional

	DB	'W'	; Wide/Normal print

	DB	'-'	; Underline on/off

	DB	'_'	; Overline on/off



	DB	'p'	; Proportional Spacing

	DB	'!'	; Pitch Select

	DB	'R'	; Int'l Character Set

	DB	'r'	; JX-80, NX-1000 Rainbow: Ribbon Select



	DB	'h'	; Select double or quadruple size

	DB	'w'	; Selsize

	DB	'w'	; Select double or single height



	DB	'a'	; Justification

	DB	't'	; IBM/Std. Character Set

	DB	'/'	; Vert Tab Channel

	DB	'%'	; ROM/Download char's

	DB	'x'	; NLQ/Draft characters

	DB	'~'	; Normal/Slash zero



	DB	0	; End of Table 4

;

;	State 5 -- ESC followed by a TWO character argument.

;

	DB	'?'	; Convert graphics density

	DB	'\'	; Horiz Tab (Star) or Print All Chars (IBM)

	DB	'$'	; Abs Horiz Tab in Inches

	DB	0	; End of Table 5



	DB	0	; End of all State Tables

 ENDIF

 IF NEC

;	NEC-8023/C.Itoh 8510

;

;	Type 1 -- Graphics

;

	DB	'S'	; Graphics

	DB	0	; (end state 1)



	DB	0	; End of all State Tables

 ENDIF

 IF MPI

;	For MPI 99 and MPI 150 Driver



;

;	Type 1 -- Graphics (Seventh bit must remain on during entire

;		  binary transfer, as per printer requirements)

;

	DB	23	; Enter graphics mode

	DB	'W'	; "

	DB	8	; Download 7 wide font

	DB	'H'	; "

	DB	6	; Download 11 wide font

	DB	'F'	; "



	DB	0	; (end state 1)



	DB	0	; End of all State Tables



 ENDIF

 IF OKI

;	For OkidataMicroline



	DB	0	; End state 1

	DB	0	; End of all State Tables

 ENDIF

 IF H25

;	For H-25 Printer

;

;	Type 1 -- Escape-OpenBracket means that the escape code is multi-

;		  character, and is terminated by character >= 'A'.

;

	DB	'['	; 



	DB	0	; (end state 1)



	DB	0	; End of all State Tables

 ENDIF

 IF DIABLO

;	Diablo-compatible Daisywheel

;

;	Type 1 -- Graphics

;

	DB	0	; (none)



;

;	Type 2 -- Send until NUL

;

	DB	0	; (none)



;

;	Type 3 -- Send 2d char, w/ ctrl conversion

;

	DB	09H,0BH,; 0CH,0DH,11H,1AH,1EH,1FH	

	DB	0	; (END state 3)



	DB	0	; END all state tables



 ENDIF

 IF PLAIN

;	Plain-Vanilla printer (no escape codes, or only ESC+one character)

	DB	0	; End state 1

	DB	0	; End of all State Tables

 ENDIF



;

;	Macro definitions

;

;	First byte is put into EXTSW.

 IF MX

;	EXP OFF; DO NOTHING

ODAT    DB      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;       EXP ON, DBL OFF, ENH OFF, COND OFF

	DB	1,33Q,106Q,33Q,110Q,22Q,0,0,0,0,0,0,0,0,0,0

;	EXP OFF, CONDENSED PRINT

	DB	1,ESC,'F',ESC,'H',17Q,0,0,0,0,0,0,0,0,0,0

;       EXP OFF, DBL OFF, ENH ON, COND OFF

	DB	0,33Q,106Q,33Q,105Q,22Q,0,0,0,0,0,0,0,0,0,0

;       EXP OFF, DBL OFF, ENH OFF, COND ON

	DB	0,33Q,106Q,33Q,110Q,17Q,0,0,0,0,0,0,0,0,0,0

 ENDIF

 IF NEC

ODAT	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;		PROPORTIONAL PRINT:

	DB	0,15,27,'"',27,'P',0,0,0,0,0,0,0,0,0,0

;		12 CPI ELITE PRINT

	DB	0,15,27,'"',27,'E',0,0,0,0,0,0,0,0,0,0

;		10 CPI ENHANCED PRINT

	DB	0,15,27,'!',27,'N',0,0,0,0,0,0,0,0,0,0

;		17.6 CPI CONDENSED PRINT

	DB	0,15,27,'"',27,'Q',0,0,0,0,0,0,0,0,0,0

 ENDIF

 IF MPI

ODAT	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;		12 CPI NORMAL PRINT

	DB	0,27,'O',27,']',15,0,0,0,0,0,0,0,0,0,0

;		10 CPI SERIF QUALITY PRINT

	DB	0,27,'O',27,'^',15,0,0,0,0,0,0,0,0,0,0

;		15 CPI CONDENSED PRINT

	DB	0,27,'O',27,'{',15,0,0,0,0,0,0,0,0,0,0

;		17 CPI CONDENSED PRINT

	DB	0,27,'O',27,'_',15,0,0,0,0,0,0,0,0,0,0

 ENDIF

 IF DP

ODAT	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;		12 CPI NORMAL PRINT

	DB	0,30,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;		16.8 CPI NORMAL PRINT

	DB	0,31,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;		6 CPI (12 CPI ELONGATED)

	DB	1,30,0,0,0,0,0,0,0,0,0,0,0,0,0,0

;		8.4 CPI (16.8 CPI ELONGATED)

	DB	1,31,0,0,0,0,0,0,0,0,0,0,0,0,0,0

 ENDIF

 IF OKI OR DIABLO OR PLAIN

ODAT	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

 ENDIF

 IF H25

; Always exit character-graphic mode on opening driver.

ODAT	DB	0,27,'[10m',0,0,0,0,0,0,0,0,0,0

;		10 cpi normal print

	DB	0,27,'[10m',27,'[0w',0,0,0,0,0,0

;		8.25 cpi normal print (expanded 16.5 cpi)

	DB	1,27,'[10m',27,'[4w',0,0,0,0,0,0

;		12 cpi normal print

	DB	0,27,'[10m',27,'[2w',0,0,0,0,0,0

;		16.5 cpi normal print

	DB	0,27,'[10m',27,'[4w',0,0,0,0,0,0

 ENDIF

;

;	Driver Variables:

;

PROGADR	DW	0	; Pointer into macro being sent or programmed

			; OR pointer to User Set variable, below:

;

; User-Settable variables(**DO NOT CHANGE ORDER**)

;

FLEN	DB	0	; Form length

PLEN	DB	0	; Page Length

LMVAL	DB	0	; Left Margin

TABSW	DB	8	; Tab flag: 0=Printer; any other=Driver.

PROGMAC	DB	0	; Macro # (1..8 for macros 0..7) 0=off.

;			If non-zero (in state 0) characters will be

;			saved into selected macro.  Used in state 4 to

;			save macro # to be sent.

; (END USER SETTABLE VARS)

FFSW	DB	0	; Form Feed on Close flag

;

CCC	DB	176Q	; Control conversion character

CCCSW	DB	0	; Currently interpreting ctrl conv.

NLSW	DB	0	; If true, skip next NL

;

LX	DB	1	; Line counter

CX	DB	1	; Column counter

;

GSTATE	DB	0	; Pre-processing state (0=normal)

STATE	DB	0	; Escape code interpreting state (0=normal)

;

GRLSB	DB	0	; Graphics count LSB,MSB

GRMSB	DB	0

;

EXTSW	DB	0	; Send Extended code each line?

HIBIT	DB	0	; Set hi-bit on characters? (=200Q if true)

;

; PRINTER INTERFACE PARAMETERS

LPINIT	DB	0	; Initialization flag

LPBAUD	DB	00CH,000H ; 9600 BAUD

;		006H,000H  19200

;		00CH,000H   9600

;		018H,000H   4800

;		030H,000H   2400

;		060H,000H   1200

;		0C0H,000H    600

;		080H,001H    300

;		000H,003H    150

; CAUTION   DO NOT CHANGE THE ORDER OF LPMASK AND LPCOMP

 IF U8250

LPMASK	DB	040Q	; DTR, PIN 20

LPCOMP	DB	040Q	; HIGH => READY

LPPORT	DB	340Q

 ENDIF

 IF U8251

LPMASK	DB	001Q	; DSR & TXRDY

LPCOMP	DB	001Q	; TXRDY HIGH => READY

LPPORT	DB	340Q

 ENDIF

 IF U8255 OR Z8255

LPMASK	DB	020Q	; PORT C, BIT 4

LPCOMP	DB	000Q	; LOW FOR READY HERE

LPPORT	DB	320Q

 ENDIF

 IF DIRECT

LPMASK	DB	200Q	; BUSY BIT

LPCOMP	DB	000Q	; HIGH BUSY, LOW READY

LPPORT	DB	320Q

 ENDIF

 IF PERCOM

LPMASK	DB	040Q	; BUSY BIT

LPCOMP	DB	000Q	; HIGH BUSY, LOW READY

LPPORT	DB	340Q

 ENDIF

 IF WIDGET

LPMASK	DB	140Q	; PORT C, BIT 4

LPCOMP	DB	100Q	; LOW READY

LPPORT	DB	040Q

 ENDIF

 IF H89PIP

LPMASK	DB	004H	; PORT C, BIT 3

LPCOMP	DB	000H	; LO READY <- 9/17/84

LPPORT	DB	320Q

 ENDIF

; FIX UP CALLING PATCH

SIGNON: POP	H	; GET RETURN ADDRESS

	PUSH	B	; SAVE THEIR CHARACTER

	PUSH	H

	INX	H

	INX	H

	MOV	E,M	; CRTOUT ADDRESS

	INX	H

	MOV	D,M

	XCHG

	SHLD	CRTAD+1 ; USE FOR SHOWING MESSAGE

	POP	H

	DCX	H

	DCX	H

	PUSH	H

	DCX	H

	DCX	H

	DCX	H

	DCX	H

	SHLD	CONAD+1 ; FIX RETURN JUMP

	LHLD	RTSAD

	XCHG

	POP	H

	MOV	M,E	; RESTORE CALLADDRESS

	INX	H

	MOV	M,D

; BROADCAST OUT PRESENCE

	LXI	H,MYMSG

MSGLP:	MOV	C,M

	INX	H

	MVI	A,'$'

	CMP	C

	JZ	MSGXIT

	PUSH	H

CRTAD:	CALL	0000H

	POP	H

	JMP	MSGLP

MSGXIT: POP	B

CONAD:	JMP	0000H	; GOTO CONOUT

MYMSG:	DB	CR,NL

	DB	'Ultimate Driver V6.0 (C)1990 Lindley Systems',CR,NL

	DB	'* '

 IF MX

	DB	'Epson/Compatible'

 ENDIF

 IF NEC

	DB	'NEC-8023'

 ENDIF

 IF DP

	DB	'Dataproducts'

 ENDIF

 IF MPI

	DB	'MPI'

 ENDIF

 IF OKI

	DB	'Okidata'

 ENDIF

 IF H25

	DB	'H-25'

 ENDIF

 IF PLAIN

	DB	'Plain'

 ENDIF

 IF DIABLO

	DB	'Diablo'

 ENDIF

	DB	' Printer',CR,NL,'* '

 IF U8250

	DB	'U-8250'

 ENDIF

 IF U8251

	DB	'U-8251'

 ENDIF

 IF U8255

	DB	'U-8255'

 ENDIF

 IF Z8255

	DB	'Z-8255'

 ENDIF

 IF DIRECT

	DB	'DIRECT'

 ENDIF

 IF PERCOM

	DB	'PERCOM'

 ENDIF

 IF WIDGET

	DB	'WIDGET'

 ENDIF

 IF H89PIP

	DB	'H89PIP'

 ENDIF

	DB	' Interface',CR,NL

	DB	'$'

	END

br